@@ -4,7 +4,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2022	PgPool Global Development Group
+ * Copyright (c) 2003-2019	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -618,13 +618,14 @@ inform_process_count(PCP_CONNECTION * frontend)
 
 	process_list = pool_get_process_list(&process_count);
 
-	mesg = (char *) palloc(8 * process_count);	/* PID is at most 7 characters long */
+	mesg = (char *) palloc(7 * process_count);	/* PID is at most 6 characters
+												 * long */
 
 	snprintf(process_count_str, sizeof(process_count_str), "%d", process_count);
 
 	for (i = 0; i < process_count; i++)
 	{
-		char		process_id[8];
+		char		process_id[7];
 
 		snprintf(process_id, sizeof(process_id), "%d", process_list[i]);
 		snprintf(mesg + total_port_len, strlen(process_id) + 1, "%s", process_id);
@@ -818,7 +819,6 @@ inform_node_info(PCP_CONNECTION * frontend, char *buf)
 	int			wsize;
 	char		port_str[6];
 	char		status[2];
-	char		quarantine[2];
 	char		weight_str[20];
 	char		role_str[10];
 	char		standby_delay_str[20];
@@ -842,7 +842,6 @@ inform_node_info(PCP_CONNECTION * frontend, char *buf)
 
 	snprintf(port_str, sizeof(port_str), "%d", bi->backend_port);
 	snprintf(status, sizeof(status), "%d", bi->backend_status);
-	snprintf(quarantine, sizeof(quarantine), "%d", bi->quarantine);
 	snprintf(weight_str, sizeof(weight_str), "%f", bi->backend_weight);
 
 	if (STREAM)
@@ -870,10 +869,11 @@ inform_node_info(PCP_CONNECTION * frontend, char *buf)
 				  strlen(bi->backend_hostname) + 1 +
 				  strlen(port_str) + 1 +
 				  strlen(status) + 1 +
-				  strlen(quarantine) + 1 +
 				  strlen(weight_str) + 1 +
 				  strlen(role_str) + 1 +
 				  strlen(standby_delay_str) + 1 +
+				  strlen(bi->replication_state) + 1 +
+				  strlen(bi->replication_sync_state) + 1 +
 				  strlen(status_changed_time_str) + 1 +
 				  sizeof(int));
 	pcp_write(frontend, &wsize, sizeof(int));
@@ -881,10 +881,11 @@ inform_node_info(PCP_CONNECTION * frontend, char *buf)
 	pcp_write(frontend, bi->backend_hostname, strlen(bi->backend_hostname) + 1);
 	pcp_write(frontend, port_str, strlen(port_str) + 1);
 	pcp_write(frontend, status, strlen(status) + 1);
-	pcp_write(frontend, quarantine, strlen(quarantine) + 1);
 	pcp_write(frontend, weight_str, strlen(weight_str) + 1);
 	pcp_write(frontend, role_str, strlen(role_str) + 1);
 	pcp_write(frontend, standby_delay_str, strlen(standby_delay_str) + 1);
+	pcp_write(frontend, bi->replication_state, strlen(bi->replication_state) + 1);
+	pcp_write(frontend, bi->replication_sync_state, strlen(bi->replication_sync_state) + 1);
 	pcp_write(frontend, status_changed_time_str, strlen(status_changed_time_str) + 1);
 
 	do_pcp_flush(frontend);
