@@ -679,7 +679,11 @@ read_startup_packet(POOL_CONNECTION * cp)
 				}
 				else
 				{
+					ereport(DEBUG1,
+							(errmsg("reading startup packet"),
+							 errdetail("guc name: %s value: %s", p, p+strlen(p)+1)));
 					p += (strlen(p) + 1);
+
 				}
 
 				p += (strlen(p) + 1);
@@ -1655,7 +1659,7 @@ select_load_balancing_node(void)
 							   pool_config->db_redirect_tokens->token[index_db].weight_token)));
 
 			tmp = choose_db_node_id(pool_config->db_redirect_tokens->token[index_db].right_token);
-			if (tmp == -1 || (tmp >= 0 && VALID_BACKEND(tmp)))
+			if (tmp == -1 || (tmp >= 0 && VALID_BACKEND_RAW(tmp)))
 				suggested_node_id = tmp;
 		}
 	}
@@ -1696,7 +1700,7 @@ select_load_balancing_node(void)
 								   pool_config->app_name_redirect_tokens->token[index_app].weight_token)));
 
 				tmp = choose_db_node_id(pool_config->app_name_redirect_tokens->token[index_app].right_token);
-				if (tmp == -1 || (tmp >= 0 && VALID_BACKEND(tmp)))
+				if (tmp == -1 || (tmp >= 0 && VALID_BACKEND_RAW(tmp)))
 					suggested_node_id = tmp;
 			}
 		}
