@@ -61,6 +61,9 @@
 #include "auth/pool_hba.h"
 #include "utils/pool_stream.h"
 
+#include "watchdog/wd_ipc_commands.h"
+#include "watchdog/watchdog.h"
+
 static POOL_CONNECTION_POOL_SLOT * slots[MAX_NUM_BACKENDS];
 static volatile sig_atomic_t reload_config_request = 0;
 static volatile sig_atomic_t restart_request = 0;
@@ -73,6 +76,7 @@ static unsigned long long int text_to_lsn(char *text);
 static RETSIGTYPE my_signal_handler(int sig);
 static RETSIGTYPE reload_config_handler(int sig);
 static void reload_config(void);
+static void sr_check_will_die(int code, Datum arg);
 
 #define CHECK_REQUEST \
 	do { \
@@ -89,6 +93,9 @@ static void reload_config(void);
 
 
 #define PG10_SERVER_VERSION	100000	/* PostgreSQL 10 server version num */
+#define PG91_SERVER_VERSION	90100	/* PostgreSQL 9.1 server version num */
+
+static	volatile bool follow_primary_lock_acquired;
 
 /*
 * worker child main loop
@@ -106,6 +113,11 @@ do_worker_child(void)
 	init_ps_display("", "", "", "");
 	set_ps_display("worker process", false);
 
+	/*
+	 * install the call back for preparation of exit
+	 */
+	on_system_exit(sr_check_will_die, (Datum) NULL);
+
 	/* set up signal handlers */
 	signal(SIGALRM, SIG_DFL);
 	signal(SIGTERM, my_signal_handler);
@@ -148,6 +160,7 @@ do_worker_child(void)
 	{
 		MemoryContextSwitchTo(WorkerMemoryContext);
 		MemoryContextResetAndDeleteChildren(WorkerMemoryContext);
+		WD_STATES	wd_status;
 
 		CHECK_REQUEST;
 
@@ -157,51 +170,91 @@ do_worker_child(void)
 		}
 
 		/*
-		 * If streaming replication mode, do time lag checking
+		 * Get watchdog status if watchdog is enabled.
 		 */
-
-		if (pool_config->sr_check_period > 0 && STREAM)
+		if (pool_config->use_watchdog)
 		{
-			establish_persistent_connection();
-			PG_TRY();
+			wd_status = get_watchdog_local_node_state();
+			ereport(DEBUG1,
+					(errmsg("watchdog status: %d", wd_status)));
+		}
+
+		/*
+		 * If streaming replication mode, do time lag checking
+		 * Also skip if failover/failback is ongoing.
+		 */
+		if (pool_config->sr_check_period > 0 && STREAM &&
+			Req_info->switching == false)
+		{
+			/*
+			 * Acquire follow primary lock. If fail to acqure lock, try again.
+			 */
+			follow_primary_lock_acquired = false;
+
+			if (pool_acquire_follow_primary_lock(false) == true)
 			{
-				POOL_NODE_STATUS *node_status;
-				int			i;
+				follow_primary_lock_acquired = true;
 
-				/* Do replication time lag checking */
-				check_replication_time_lag();
-
-				/* Check node status */
-				node_status = verify_backend_node_status(slots);
-				for (i = 0; i < NUM_BACKENDS; i++)
+				establish_persistent_connection();
+				PG_TRY();
 				{
-					ereport(DEBUG1,
-							(errmsg("node status[%d]: %d", i, node_status[i])));
+					POOL_NODE_STATUS *node_status;
+					int			i;
 
-					if (node_status[i] == POOL_NODE_STATUS_INVALID)
+					/* Do replication time lag checking */
+					check_replication_time_lag();
+
+					/* Check node status */
+					node_status = verify_backend_node_status(slots);
+
+
+					for (i = 0; i < NUM_BACKENDS; i++)
 					{
-						int			n;
+						ereport(DEBUG1,
+								(errmsg("node status[%d]: %d", i, node_status[i])));
 
-						ereport(LOG,
-								(errmsg("pgpool_worker_child: invalid node found %d", i)));
-						if (pool_config->detach_false_primary)
+						if (node_status[i] == POOL_NODE_STATUS_INVALID)
 						{
-							n = i;
-							degenerate_backend_set(&n, 1, REQ_DETAIL_SWITCHOVER | REQ_DETAIL_CONFIRMED);
+							int			n;
+
+							ereport(LOG,
+									(errmsg("pgpool_worker_child: invalid node found %d", i)));
+							/*
+							 * If detach_false_primary is enabled, send
+							 * degenerate request to detach invalid node.
+							 * This should only happen on leader watchdog node
+							 * and quorum exists if watchdog is enabled. Other
+							 * nodes will be informed by the leader node later
+							 * on.
+							 */
+							if ((pool_config->detach_false_primary && !pool_config->use_watchdog) ||
+								(pool_config->detach_false_primary && pool_config->use_watchdog &&
+								 get_watchdog_quorum_state() >= 0 && wd_status == WD_COORDINATOR))
+							{
+								n = i;
+								degenerate_backend_set(&n, 1, REQ_DETAIL_SWITCHOVER | REQ_DETAIL_CONFIRMED);
+							}
 						}
 					}
 				}
-			}
-			PG_CATCH();
-			{
-				discard_persistent_connection();
-				sleep(pool_config->sr_check_period);
-				PG_RE_THROW();
-			}
-			PG_END_TRY();
+				PG_CATCH();
+				{
+					discard_persistent_connection();
+					pool_release_follow_primary_lock();
+					follow_primary_lock_acquired = false;
+					sleep(pool_config->sr_check_period);
+					PG_RE_THROW();
+				}
+				PG_END_TRY();
 
-			/* Discard persistent connections */
-			discard_persistent_connection();
+				/* Discard persistent connections */
+				discard_persistent_connection();
+				if (follow_primary_lock_acquired)
+				{
+					pool_release_follow_primary_lock();
+					follow_primary_lock_acquired = false;
+				}
+			}
 		}
 		sleep(pool_config->sr_check_period);
 	}
@@ -232,7 +285,7 @@ establish_persistent_connection(void)
 															 bkinfo->backend_port,
 															 pool_config->sr_check_database,
 															 pool_config->sr_check_user,
-															 password ? password : "", false);
+															 password ? password : "", true);
 		}
 	}
 
@@ -268,13 +321,24 @@ check_replication_time_lag(void)
 	static int	server_version[MAX_NUM_BACKENDS];
 
 	int			i;
-	int			active_nodes = 0;
 	POOL_SELECT_RESULT *res;
+	POOL_SELECT_RESULT *res_rep;	/* query results of pg_stat_replication */
 	unsigned long long int lsn[MAX_NUM_BACKENDS];
 	char	   *query;
+	char	   *stat_rep_query;
 	BackendInfo *bkinfo;
 	unsigned long long int lag;
 	ErrorContextCallback callback;
+	int		active_standby_node;
+
+	/* clear replication state */
+	for (i = 0; i < NUM_BACKENDS; i++)
+	{
+		bkinfo = pool_get_node_info(i);
+
+		*bkinfo->replication_state = '\0';
+		*bkinfo->replication_sync_state = '\0';
+	}
 
 	if (NUM_BACKENDS <= 1)
 	{
@@ -288,36 +352,6 @@ check_replication_time_lag(void)
 		return;
 	}
 
-	/* Count healthy nodes */
-	for (i = 0; i < NUM_BACKENDS; i++)
-	{
-		if (VALID_BACKEND(i))
-			active_nodes++;
-	}
-
-	if (active_nodes <= 1)
-	{
-		/*
-		 * If there's only one or less active node, there's no point to do
-		 * checking
-		 */
-		return;
-	}
-
-	if (!VALID_BACKEND(REAL_PRIMARY_NODE_ID))
-	{
-		/*
-		 * No need to check replication delay if primary is down. This could
-		 * happen if ALWAYS_PRIMARY flag is on because REAL_PRIMARY_NODE_ID
-		 * macro returns the node id which ALWAYS_PRIMARY flag is set to.  If
-		 * we do not check this, subsequent test (i == PRIMARY_NODE_ID) in the
-		 * for loop below will return unexpected result because
-		 * PRIMARY_NODE_ID macro returns MAIN_NODE_ID, which could be a
-		 * standby server.
-		 */
-		return;
-	}
-
 	/*
 	 * Register a error context callback to throw proper context message
 	 */
@@ -325,6 +359,8 @@ check_replication_time_lag(void)
 	callback.arg = NULL;
 	callback.previous = error_context_stack;
 	error_context_stack = &callback;
+	stat_rep_query = NULL;
+	active_standby_node = 0;
 
 	for (i = 0; i < NUM_BACKENDS; i++)
 	{
@@ -344,7 +380,7 @@ check_replication_time_lag(void)
 
 		if (server_version[i] == 0)
 		{
-			query = "SELECT pg_catalog.current_setting('server_version_num')";
+			query = "SELECT current_setting('server_version_num')";
 
 			/*
 			 * Get backend server version. If the query fails, keep previous
@@ -362,16 +398,23 @@ check_replication_time_lag(void)
 		if (PRIMARY_NODE_ID == i)
 		{
 			if (server_version[i] >= PG10_SERVER_VERSION)
-				query = "SELECT pg_catalog.pg_current_wal_lsn()";
+				query = "SELECT pg_current_wal_lsn()";
 			else
-				query = "SELECT pg_catalog.pg_current_xlog_location()";
+				query = "SELECT pg_current_xlog_location()";
+
+			if (server_version[i] == PG91_SERVER_VERSION)
+				stat_rep_query = "SELECT application_name, state, '' AS sync_state FROM pg_stat_replication";
+			else if (server_version[i] > PG91_SERVER_VERSION)
+				stat_rep_query = "SELECT application_name, state, sync_state FROM pg_stat_replication";
 		}
 		else
 		{
 			if (server_version[i] >= PG10_SERVER_VERSION)
-				query = "SELECT pg_catalog.pg_last_wal_replay_lsn()";
+				query = "SELECT pg_last_wal_replay_lsn()";
 			else
-				query = "SELECT pg_catalog.pg_last_xlog_replay_location()";
+				query = "SELECT pg_last_xlog_replay_location()";
+
+			active_standby_node++;
 		}
 
 		if (get_query_result(slots, i, query, &res) == 0 && res->nullflags[0] != -1)
@@ -381,6 +424,58 @@ check_replication_time_lag(void)
 		}
 	}
 
+	/*
+	 * Call pg_stat_replication and fill the replication status.
+	 */
+	if (slots[PRIMARY_NODE_ID] && stat_rep_query != NULL)
+	{
+		int		status;
+
+		status = get_query_result(slots, PRIMARY_NODE_ID, stat_rep_query, &res_rep);
+
+		if (status == -1 || (status == -2 && active_standby_node > 0))
+		{
+			ereport(LOG,
+					(errmsg("get_query_result falied: status: %d", status)));
+		}
+
+		for (i = 0; i < NUM_BACKENDS; i++)
+		{
+			bkinfo = pool_get_node_info(i);
+
+			*bkinfo->replication_state = '\0';
+			*bkinfo->replication_sync_state = '\0';
+
+			if (i == PRIMARY_NODE_ID)
+				continue;
+
+			if (status == 0)
+			{
+				int		j;
+				char	*s;
+
+				for (j = 0; j < res_rep->numrows; j++)
+				{
+					if (strcmp(res_rep->data[j*3], bkinfo->backend_application_name) == 0)
+					{
+						/*
+						 * If sr_check_user has enough privilege, it should return
+						 * some string. If not, NULL pointer will be returned for
+						 * res_rep->data[1] and [2]. So we need to prepare for the
+						 * latter case.
+						 */
+						s = res_rep->data[j*3+1]? res_rep->data[j*3+1] : "";
+						strlcpy(bkinfo->replication_state, s, NAMEDATALEN);
+						s = res_rep->data[j*3+2]? res_rep->data[j*3+2] : "";
+						strlcpy(bkinfo->replication_sync_state, s, NAMEDATALEN);
+					}
+				}
+			}
+		}
+		if (status == 0)
+			free_select_result(res_rep);
+	}
+
 	for (i = 0; i < NUM_BACKENDS; i++)
 	{
 		if (!VALID_BACKEND(i))
@@ -506,9 +601,9 @@ reload_config(void)
 
 /*
  * Execute query against specified backend using an established connection to
- * backend.  Return -1 on failure or 0 otherwise.  Caller must prepare memory
- * for POOL_SELECT_RESULT and pass it as "res". It is guaranteed that no
- * exception occurs within this function.
+ * backend.  Return -1 on failure, -2 on no rows returned, or 0 otherwise.
+ * Caller must prepare memory for POOL_SELECT_RESULT and pass it as "res". It
+ * is guaranteed that no exception occurs within this function.
  */
 int
 get_query_result(POOL_CONNECTION_POOL_SLOT * *slots, int backend_id, char *query, POOL_SELECT_RESULT * *res)
@@ -542,32 +637,20 @@ get_query_result(POOL_CONNECTION_POOL_SLOT * *slots, int backend_id, char *query
 	if ((*res)->numrows <= 0)
 	{
 		free_select_result(*res);
-		ereport(LOG,
+		ereport(DEBUG1,
 				(errmsg("get_query_result: no rows returned"),
 				 errdetail("node id (%d)", backend_id)));
-		return sts;
+		return -2;
 	}
 
-/*
-	if ((*res)->data[0] == NULL)
-	{
-		free_select_result(*res);
-		ereport(LOG,
-				(errmsg("get_query_result: no rows returned"),
-				 errdetail("node id (%d)", backend_id)));
-		return sts;
-	}
-
-
-	if ((*res)->nullflags[0] == -1)
-	{
-		free_select_result(*res);
-		ereport(LOG,
-				(errmsg("get_query_result: NULL data returned"),
-				 errdetail("node id (%d)", backend_id)));
-		return sts;
-	}
-*/
 	sts = 0;
 	return sts;
 }
+
+static void
+sr_check_will_die(int code, Datum arg)
+{
+	if (follow_primary_lock_acquired)
+		pool_release_follow_primary_lock();
+
+}
