@@ -3,7 +3,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2023	PgPool Global Development Group
+ * Copyright (c) 2003-2021	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -353,9 +353,7 @@ function_call_walker(Node *node, void *context)
 
 			if (ctx->pg_terminate_backend_pid == 0 && strcmp("pg_terminate_backend", fname) == 0)
 			{
-				/* PostgreSQL 14 or after accept two arguments. */
-				if (list_length(fcall->args) == 1 ||
-					list_length(fcall->args) == 2)
+				if (list_length(fcall->args) == 1)
 				{
 					Node	   *arg = linitial(fcall->args);
 
@@ -525,8 +523,7 @@ view_walker(Node *node, void *context)
 }
 
 /*
- * Judge the table used in a query represented by node is a system
- * catalog or not.
+ * Determine whether table_name is a system catalog or not.
  */
 static bool
 is_system_catalog(char *table_name)
@@ -539,15 +536,13 @@ is_system_catalog(char *table_name)
 /*
  * Query to know if the target table belongs pg_catalog schema.
  */
-#define ISBELONGTOPGCATALOGQUERY "SELECT count(*) FROM pg_catalog.pg_class AS c, pg_catalog.pg_namespace AS n WHERE c.relname = '%s' AND c.relnamespace = n.oid AND n.nspname = 'pg_catalog'"
+#define ISBELONGTOPGCATALOGQUERY "SELECT count(*) FROM pg_class AS c, pg_namespace AS n WHERE c.relname = '%s' AND c.relnamespace = n.oid AND n.nspname = 'pg_catalog'"
 
-#define ISBELONGTOPGCATALOGQUERY2 "SELECT count(*) FROM pg_catalog.pg_class AS c, pg_catalog.pg_namespace AS n WHERE c.oid = pgpool_regclass('\"%s\"') AND c.relnamespace = n.oid AND n.nspname = 'pg_catalog'"
+#define ISBELONGTOPGCATALOGQUERY2 "SELECT count(*) FROM pg_class AS c, pg_namespace AS n WHERE c.oid = pgpool_regclass('\"%s\"') AND c.relnamespace = n.oid AND n.nspname = 'pg_catalog'"
 
-#define ISBELONGTOPGCATALOGQUERY3 "SELECT count(*) FROM pg_catalog.pg_class AS c, pg_catalog.pg_namespace AS n WHERE c.oid = pg_catalog.to_regclass('\"%s\"') AND c.relnamespace = n.oid AND n.nspname = 'pg_catalog'"
+#define ISBELONGTOPGCATALOGQUERY3 "SELECT count(*) FROM pg_class AS c, pg_namespace AS n WHERE c.oid = pg_catalog.to_regclass('\"%s\"') AND c.relnamespace = n.oid AND n.nspname = 'pg_catalog'"
 
-	int			hasreliscatalog;
 	bool		result;
-	static POOL_RELCACHE * hasreliscatalog_cache;
 	static POOL_RELCACHE * relcache;
 	POOL_CONNECTION_POOL *backend;
 
@@ -562,28 +557,9 @@ is_system_catalog(char *table_name)
 	backend = pool_get_session_context(false)->backend;
 
 	/*
-	 * Check if pg_namespace exists
+	 * Check if pg_namespace exists. PostgreSQL 7.3 or later has it.
 	 */
-	if (!hasreliscatalog_cache)
-	{
-		char	   *query;
-
-		query = HASPGNAMESPACEQUERY;
-
-		hasreliscatalog_cache = pool_create_relcache(pool_config->relcache_size, query,
-													 int_register_func, int_unregister_func,
-													 false);
-		if (hasreliscatalog_cache == NULL)
-		{
-			ereport(WARNING,
-					(errmsg("unable to create relcache, while checking for system catalog")));
-			return false;
-		}
-	}
-
-	hasreliscatalog = pool_search_relcache(hasreliscatalog_cache, backend, "pg_namespace") == 0 ? 0 : 1;
-
-	if (hasreliscatalog)
+	if (Pgversion(backend)->major >= 73)
 	{
 		/*
 		 * If relcache does not exist, create it.
@@ -632,8 +608,7 @@ is_system_catalog(char *table_name)
 }
 
 /*
- * Judge the table used in a query represented by node is a temporary
- * table or not.
+ * Returns true if table_name is a temporary table.
  */
 static POOL_RELCACHE * is_temp_table_relcache;
 
@@ -652,7 +627,7 @@ is_temp_table(char *table_name)
  * regclass (or its variant) here, because temporary tables never have
  * schema qualified name.
  */
-#define ISTEMPQUERY83 "SELECT count(*) FROM pg_catalog.pg_class AS c, pg_catalog.pg_namespace AS n WHERE c.relname = '%s' AND c.relnamespace = n.oid AND n.nspname ~ '^pg_temp_'"
+#define ISTEMPQUERY83 "SELECT count(*) FROM pg_catalog.pg_class AS c, pg_namespace AS n WHERE c.relname = '%s' AND c.relnamespace = n.oid AND n.nspname ~ '^pg_temp_'"
 
 /*
  * Query to know if the target table is a temporary one.  This query
@@ -661,41 +636,57 @@ is_temp_table(char *table_name)
  * name.
  */
 #define ISTEMPQUERY84 "SELECT count(*) FROM pg_catalog.pg_class AS c WHERE c.relname = '%s' AND c.relistemp"
-
-	int			hasrelistemp;
 	bool		result;
-	static POOL_RELCACHE * hasrelistemp_cache;
 	char	   *query;
 	POOL_CONNECTION_POOL *backend;
+	int			major;
 
-	if (table_name == NULL)
+	if (table_name == NULL || pool_config->check_temp_table == CHECK_TEMP_NONE ||
+		pool_config->check_temp_table == CHECK_TEMP_OFF)
 	{
 		return false;
 	}
 
 	backend = pool_get_session_context(false)->backend;
 
-	/*
-	 * Check backend version
-	 */
-	if (!hasrelistemp_cache)
+	if (pool_config->check_temp_table == CHECK_TEMP_TRACE)
 	{
-		hasrelistemp_cache = pool_create_relcache(pool_config->relcache_size, HASRELITEMPPQUERY,
-												  int_register_func, int_unregister_func,
-												  false);
-		if (hasrelistemp_cache == NULL)
+		POOL_TEMP_TABLE	*temp_table;
+
+		temp_table = pool_temp_tables_find(table_name);
+		if (temp_table && (temp_table->state == TEMP_TABLE_CREATE_COMMITTED ||
+						   temp_table->state == TEMP_TABLE_CREATING))
 		{
-			ereport(WARNING,
-					(errmsg("unable to create relcache, while checking for temporary table")));
+			ereport(DEBUG1,
+					(errmsg("is_temp_table: %s is a temp table", table_name)));
+			return true;
+		}
+		else
+		{
+			ereport(DEBUG1,
+					(errmsg("is_temp_table: %s is not a temp table", table_name)));
 			return false;
 		}
 	}
 
-	hasrelistemp = pool_search_relcache(hasrelistemp_cache, backend, "pg_class") == 0 ? 0 : 1;
-	if (hasrelistemp)
+	/*
+	 * Below is check_temp_table == CHECK_TEMP_CATALOG or CHECK_TEMP_ON case.
+	 */
+
+	/*
+	 * Check backend version.
+	 */
+	major = Pgversion(backend)->major;
+	if (major == 73 || major == 83 || major >= 91)
+		query = ISTEMPQUERY83;
+	else if (major == 84 || major == 90)
 		query = ISTEMPQUERY84;
 	else
-		query = ISTEMPQUERY83;
+	{
+		ereport(WARNING,
+				(errmsg("is_temp_table: unexpected PostgreSQL version: %s", Pgversion(backend)->version_string)));
+		return false;	/* fall back to assume that the table is not a temporary table. */
+	}
 
 	/*
 	 * If relcache does not exist, create it.
@@ -734,8 +725,7 @@ discard_temp_table_relcache(void)
 }
 
 /*
- * Judge the table used in a query represented by node is a unlogged
- * table or not.
+ * Returns true if table_name is an unlogged table.
  */
 bool
 is_unlogged_table(char *table_name)
@@ -756,10 +746,9 @@ is_unlogged_table(char *table_name)
 
 #define ISUNLOGGEDQUERY3 "SELECT count(*) FROM pg_catalog.pg_class AS c WHERE c.oid = pg_catalog.to_regclass('%s') AND c.relpersistence = 'u'"
 
-	int			hasrelpersistence;
-	static POOL_RELCACHE * hasrelpersistence_cache;
 	static POOL_RELCACHE * relcache;
 	POOL_CONNECTION_POOL *backend;
+	int		major;
 
 	if (table_name == NULL)
 	{
@@ -772,23 +761,10 @@ is_unlogged_table(char *table_name)
 	backend = pool_get_session_context(false)->backend;
 
 	/*
-	 * Check backend version
+	 * Check backend version.
 	 */
-	if (!hasrelpersistence_cache)
-	{
-		hasrelpersistence_cache = pool_create_relcache(pool_config->relcache_size, HASRELPERSISTENCEQUERY,
-													   int_register_func, int_unregister_func,
-													   false);
-		if (hasrelpersistence_cache == NULL)
-		{
-			ereport(WARNING,
-					(errmsg("unable to create relcache, while checking for unlogged table")));
-			return false;
-		}
-	}
-
-	hasrelpersistence = pool_search_relcache(hasrelpersistence_cache, backend, "pg_class") == 0 ? 0 : 1;
-	if (hasrelpersistence)
+	major = Pgversion(backend)->major;
+	if (major >= 91)
 	{
 		bool		result;
 		char	   *query;
@@ -837,7 +813,8 @@ is_unlogged_table(char *table_name)
 }
 
 /*
- * Judge the table used in a query is a view or not.
+ * Returns true if table_name is a view.
+ * This function is called by query cache module.
  */
 bool
 is_view(char *table_name)
@@ -911,7 +888,7 @@ pool_has_pgpool_regclass(void)
 /*
  * Query to know if pgpool_regclass exists.
  */
-#define HASPGPOOL_REGCLASSQUERY "SELECT count(*) from (SELECT pg_catalog.has_function_privilege('%s', 'pgpool_regclass(cstring)', 'execute') WHERE EXISTS(SELECT * FROM pg_catalog.pg_proc AS p WHERE p.proname = 'pgpool_regclass')) AS s"
+#define HASPGPOOL_REGCLASSQUERY "SELECT count(*) from (SELECT has_function_privilege('%s', 'pgpool_regclass(cstring)', 'execute') WHERE EXISTS(SELECT * FROM pg_catalog.pg_proc AS p WHERE p.proname = 'pgpool_regclass')) AS s"
 
 	bool		result;
 	static POOL_RELCACHE * relcache;
@@ -939,39 +916,22 @@ pool_has_pgpool_regclass(void)
 }
 
 /*
- * Judge if we have to_regclass or not.
+ * Returns true if we have to_regclass().
  */
 bool
 pool_has_to_regclass(void)
 {
-/*
- * Query to know if to_regclass exists.
- */
-#define HAS_TOREGCLASSQUERY "SELECT count(*) from (SELECT has_function_privilege('%s', 'pg_catalog.to_regclass(cstring)', 'execute') WHERE EXISTS(SELECT * FROM pg_catalog.pg_proc AS p WHERE p.proname = 'to_regclass')) AS s"
-
-	bool		result;
-	static POOL_RELCACHE * relcache;
 	POOL_CONNECTION_POOL *backend;
-	char	   *user;
+	PGVersion	*pgversion;
 
 	backend = pool_get_session_context(false)->backend;
-	user = MASTER_CONNECTION(backend)->sp->user;
+	pgversion = Pgversion(backend);
 
-	if (!relcache)
-	{
-		relcache = pool_create_relcache(pool_config->relcache_size, HAS_TOREGCLASSQUERY,
-										int_register_func, int_unregister_func,
-										false);
-		if (relcache == NULL)
-		{
-			ereport(WARNING,
-					(errmsg("unable to create relcache, while checking to_regclass presence")));
-			return false;
-		}
-	}
+	/* PostgreSQL 9.4 or above has to_regclass() */
+	if (pgversion->major >= 94)
+		return true;
 
-	result = pool_search_relcache(relcache, backend, user) == 0 ? 0 : 1;
-	return result;
+	return false;
 }
 
 /*
@@ -1056,22 +1016,16 @@ non_immutable_function_call_walker(Node *node, void *context)
 			}
 		}
 	}
-
-	/* Before Pgpool-II 3.7 there was no SQLValueFunction in the parser.  For
-	 * the older versions (3.6 or before), we need to check type cast and
-	 * typename instead.  If they are some type casts described below, we
-	 * assume that this SELECT cannot be cached since they might be a
-	 * transformed CURRENT_TIMESTAMP etc.  Of course this could be overkill as
-	 * "SELECT '2022-07-04 09:00:00'::TIMESTAMP" could be regarded as
-	 * non-cachable for example. But there's nothing we can do here.
-	 */
 	else if (IsA(node, TypeCast))
 	{
-		/* TIMESTAMP WITH TIME ZONE and TIME WITH TIME ZONE should not be cached. */
+		/* CURRENT_DATE, CURRENT_TIME, LOCALTIMESTAMP, LOCALTIME etc. */
 		TypeCast   *tc = (TypeCast *) node;
 
-		if (isSystemType((Node *) tc->typeName, "timestamptz") ||
-			isSystemType((Node *) tc->typeName, "timetz"))
+		if ((isSystemType((Node *) tc->typeName, "date") ||
+			 isSystemType((Node *) tc->typeName, "timestamp") ||
+			 isSystemType((Node *) tc->typeName, "timestamptz") ||
+			 isSystemType((Node *) tc->typeName, "time") ||
+			 isSystemType((Node *) tc->typeName, "timetz")))
 		{
 			ctx->has_non_immutable_function_call = true;
 			return false;
@@ -1140,7 +1094,7 @@ pool_table_name_to_oid(char *table_name)
  * Query to convert table name to oid
  */
 #define TABLE_TO_OID_QUERY "SELECT pgpool_regclass('%s')"
-#define TABLE_TO_OID_QUERY2 "SELECT oid FROM pg_catalog.pg_class WHERE relname = '%s'"
+#define TABLE_TO_OID_QUERY2 "SELECT oid FROM pg_class WHERE relname = '%s'"
 #define TABLE_TO_OID_QUERY3 "SELECT COALESCE(pg_catalog.to_regclass('%s')::oid, 0)"
 
 	int			oid = 0;
